#include "Essentials.h"

typedef struct HidingLink HidingLink; /// for keeping nested 'hiding's

struct HidingLink {
	HidingLink		*next;
	Error			*keepError;
};

typedef struct ContextLink ContextLink; /// for keeping nested contexts

struct ContextLink {
	ContextLink		*next;
	Str255			contextString;
};

HidingLink			*firstHiding=0; /// first one is most recently added
ContextLink			*firstContext=0; /// first one is most recently added

Boolean HideErrors(Error* keepError)
{
	Error			error;
	HidingLink		*link;

	Assert(keepError);
	
	link=(HidingLink*)AllocPtr(&error, sizeof(HidingLink)); /// Create hiding link
		_i(error) /// Don't report the error, of course!
	link->keepError=keepError; /// Assign the keep error pointer
	*keepError=0; /// Initialise error to zero (in case it's never set again)
	
	AddToLinkedList((void**)&firstHiding, link, firstHiding); /// Add to start of list
	
	return true;
	_e
	return false;
}

void ShowErrors()
{
	HidingLink		*link;
	
	link=firstHiding; /// Get the most recently added link
	Assert(link); /// Check we haven't unnested too far
	RemoveFromLinkedList((void**)&firstHiding, link); /// Remove the link
	DestroyPtr(link); /// Free up memory for the link	
}

Boolean AddErrorContext(Str255 contextString)
{
	Error		error;
	ContextLink	*link;
	
	Assert(contextString);
	
	link=(ContextLink*)AllocPtr(&error, sizeof(ContextLink)); /// Create context link
		_i(error) /// Don't report the error, of course!
	BlockMove(contextString, link->contextString, 1+*contextString); /// Copy the string
	
	AddToLinkedList((void**)&firstContext, link, firstContext); /// Add to start of list
	
	return true;
	_e
	return false;
}

void RemoveErrorContext()
{
	ContextLink		*link;
	
	link=firstContext; /// Get the most recently added link
	Assert(link); /// Check we haven't unnested too far
	RemoveFromLinkedList((void**)&firstContext, link); /// Remove the link
	DestroyPtr(link); /// Free up memory for the link	
}

void TestError(Error error)
{
	Str255			contextString, IDString, errorString, solutionString;
	short			numString, totalStrings, compareError, resource, resources;
	Handle			errorsResource;
	unsigned char	*errorStrings, appendLen;
	HidingLink		*hiding;
	ContextLink		*context;
	
	if (error) {
	
        if (firstHiding) { /// If we're hiding errors
			hiding=firstHiding; /// Begin with first hiding
			while (hiding) {
				*(hiding->keepError)=error; /// Copy error value in
				hiding=hiding->next; /// Search for next hiding
			}
			
			SysBeep(1); /// Give a beep in any case
			
		} else {
			NumToString(error, IDString); /// Convert error code to string
			
			*errorString=0; /// To begin with, set messageString to empty
			*solutionString=0; /// To begin with, set solution string to empty
				
			resources=CountResources('Err2'); /// Count how many Err2 resources there are
			for (resource=1; resource<=resources; resource++) {

				errorsResource=GetIndResource('Err2', resource); /// Load resource
				if (ResError() || errorsResource==0)
					continue; /// If that failed, then try next one
			
				errorStrings=(unsigned char*)*errorsResource; /// Set up pointer
				totalStrings=*((short*)errorStrings)++; /// How many error strings are in resource
				
				for (numString=0; numString<totalStrings; numString++) { /// For all strings
					compareError=*((short*)errorStrings)++; /// Get error code for that string
					
					if (compareError==error) /// If this is the correct string
						BlockMove(errorStrings, errorString, *errorStrings+1); /// Copy string across
					
					errorStrings+=*errorStrings+1; /// Skip over error string
					
					if (compareError==error) {
						BlockMove(errorStrings, solutionString, *errorStrings+1); /// Copy string across
						goto stringFound; /// Need not keep searching
					}
						
					errorStrings+=*errorStrings+1; /// Skip over description string
				}
			}

			stringFound:
			if (*errorString==0)
				GetIndString(errorString, 32765, 1); /// Get "not available" string
			if (*solutionString==0)
				GetIndString(solutionString, 32765, 1); /// Get "not available" string
			
			*contextString=0; /// Begin with empty context string
			context=firstContext; /// Get first context
			while (context) {
				appendLen=*(context->contextString); /// Length of string to append to context
				if (((int)*contextString+(int)appendLen)>255) /// if the combined length is too long
					appendLen=(unsigned char)(255-*contextString); /// then truncate append length
				
				BlockMove(context->contextString+1, contextString+*contextString+1, appendLen); /// copy new data in
				*contextString+=appendLen; /// increment string's length byte
				
				context=context->next;
			}
			
			ParamText(contextString, IDString, errorString, solutionString); /// Set up dialog parameters

			Alert(32766, 0); /// Display error alert
		}
	}
}

Error TestResError(void* resource)
{
	Error	error;
	
	error=ResError(); /// First check ResError
	if (!error)
		if (resource==0L) /// Otherwise check for null handle
			error=resNotFound; /// It is unclear which error it should be, so this will do
	TestError(error); /// Call TestError in any case
	
	return error;
}

Error TestMemError(void* pointer)
{
	Error	error;
	
	error=MemError(); /// First check MemError
	if (!error)
		if (pointer==0L) /// Otherwise check for null pointer
			error=memFullErr; /// It is unclear which error it should be, so this will do
	TestError(error); /// Call TestError in any case
	
	return error;
}

#if Project_Under_Development

void HandleAssertFailure(char* file, char* date, int line)
{
	Str255			fileString, dateString, lineString, memoryString;
	unsigned char	*stringPtr;

	stringPtr=fileString+1; /// Set up copying pointer
	while (*file)
		*stringPtr++=(unsigned char)*file++; /// Terse statement to copy string across
	*fileString=(unsigned char)(stringPtr-fileString-1); /// Set length byte of fileString
	
	stringPtr=dateString+1; /// Set up copying pointer
	while (*date)
		*stringPtr++=(unsigned char)*date++; /// Terse statement to copy string across
	*dateString=(unsigned char)(stringPtr-dateString-1); /// Set length byte of dateString
	
	NumToString(line, lineString); /// Convert line number to string
	NumToString(MaxBlock(), memoryString); /// Convert Max Block available to string
	
	ParamText(fileString, lineString, dateString, memoryString); /// Set up dialog parameters

	switch (Alert(32767, 0)) { /// Determine what to do based on alert response
		case 1:
			Debugger(); /// Enter debugger
			break;
		case 2:
			ExitToShell(); /// Force quit, losing all data
			break;
		case 3:
            /// this is not really possible under Carbon Event loop
			break;
		case 4: /// Continue option leaves nothing to do
			break;
	}
}

#endif

#if Project_Under_Development

typedef struct PointerRecord PointerRecord;

struct PointerRecord {
	PointerRecord		*next;
	void				*pointer;
	OSType				signature;
};

static PointerRecord	*firstPointer=0;

void RegisterPointer(void* pointer, OSType signature)
{
	PointerRecord	*record;
	
	record=(PointerRecord*)NewPtr(sizeof(PointerRecord)); /// Allocate, don't go through AllocPtr of course!
	if (!record)
		Assert(false); /// Since this is for development only, fail to deal with lack of memory
	record->pointer=pointer; /// Place pointer inside structure
	record->signature=signature; /// Place signature inside structure
	AddToLinkedList((void**)&firstPointer, record, firstPointer); /// Add to start of record list
}

void UnregisterPointer(void* pointer, OSType signature)
{
	PointerRecord	*record;
	
	for (record=firstPointer; ; record=record->next) {
		if (!record)
			Assert(false); /// We didn't find the pointer we're disposing so fail assertion
		
		if ((record->pointer==pointer)&&(record->signature==signature)) {
			RemoveFromLinkedList((void**)&firstPointer, record); /// Remove it from the list
			DisposePtr((Ptr)record); /// Dispose the pointer we're holding it in
			break;
		}
	}
}

Boolean PointerExists(void* pointer, OSType signature)
{
	PointerRecord	*record;
	
	for (record=firstPointer; record; record=record->next)
		if ((record->pointer==pointer)&&(record->signature==signature))
			return true;
	
	return false;
}

#endif

void AddToLinkedList(void** firstPointer, void* addItem, void* beforeItem)
{
	void	**slotPointer;

	Assert(firstPointer);
	Assert(addItem);
	
	slotPointer=firstPointer;
	while (1) {

		if (*slotPointer==beforeItem) { /// See if in correct place - 0 means end
			*(void**)addItem=beforeItem; /// Make beforeItem after addItem
			*slotPointer=addItem; /// Insert addItem where beforeItem was
			break; /// Added successfully
		}
		
		Assert(*slotPointer); /// This will fail if we got to end of list
		if (*slotPointer==0)
			break; /// Even though this should never happen, fail gracefully
		slotPointer=(void**)*slotPointer; /// Get pointer to next slot
	}
}

void RemoveFromLinkedList(void** firstPointer, void* removeItem)
{
	void	**slotPointer;

	Assert(firstPointer);
	Assert(removeItem);

	slotPointer=firstPointer;
	while (1) {

		Assert(*slotPointer); /// This will fail if we got to end of list
		if (*slotPointer==0)
			break; /// Even though this should never happen, fail gracefully

		if (*slotPointer==removeItem) { /// Check if we've found our item
			*slotPointer=*(void**)removeItem; /// Put removeItem's next into its old slot
			break; /// Removed successfully
		}
		
		slotPointer=(void**)*slotPointer; /// Get pointer to next slot
	}
}

void* GetLinkedListItem(void* firstPointer, long itemNumber)
{
	void	*itemPointer;
	
	itemPointer=firstPointer;
	while(itemNumber) { /// While we're still looking for next
	
		itemPointer=*(void**)itemPointer;
		Assert(itemPointer); /// This will fail if we got to end of list
		if (itemPointer==0)
			break; /// Even though this should never happen, fail gracefully

		itemNumber--; /// Decrement number left to find
	}
	
	return itemPointer;
}

long CountLinkedListItems(void* firstPointer)
{
	void	*itemPointer;
	long	count;
	
	count=0;
	itemPointer=firstPointer;
	while (itemPointer) { /// While we're not at the end
		count++; /// Increment the count
		itemPointer=*(void**)itemPointer; /// Get to next item
	}
	
	return count;
}

#if Project_Under_Development && Initialise_Allocated_Memory
	static void InitialiseMemory(char* pointer, Size length);
#endif

void* AllocPtr(Error* error, Size allocate)
{
	void*			pointer;
	
	Assert(error);
	
	pointer=(void*)NewPtr(allocate); /// Allocate the memory
	*error=TestMemError(pointer); /// There still might have been a problem, so check
		_i(*error);
		
	#if Project_Under_Development && Initialise_Allocated_Memory
		InitialiseMemory((char*)pointer, allocate); /// Fill new block with rubbish
	#endif	

	#if Project_Under_Development && Track_Memory_Allocations
		RegisterPointer(pointer, '_ptr');
	#endif	

	return pointer;
	_e
	return 0; /// Although failure should be detectable through error, this is an extra signal
}

void** AllocHandle(Error* error, Size allocate)
{
	void**	handle;
	
	Assert(error);

	handle=(void**)NewHandle(allocate); /// Allocate the movable memory
	*error=TestMemError(handle); /// There still might have been a problem, so check
		_i(*error);
		
	#if Project_Under_Development && Move_On_Handle_Allocation
		MoveHHi((Handle)handle); /// This will probably move a lot of other blocks
	#endif

	#if Project_Under_Development && Initialise_Allocated_Memory
		InitialiseMemory((char*)*handle, allocate); /// Fill new block with rubbish
	#endif

	#if Project_Under_Development && Track_Memory_Allocations
		RegisterPointer(pointer, '_hdl');
	#endif	

	return handle;
	_e
	return 0; /// Although failure should be detectable through error, this is an extra signal
}

void DestroyPtr(void* pointer)
{
	#if Project_Under_Development && Track_Memory_Allocations
		UnregisterPointer(pointer, '_ptr');
	#endif			
			
	#if Project_Under_Development && Initialise_Allocated_Memory
		InitialiseMemory((char*)pointer, GetPtrSize((Ptr)pointer)); /// Fill with rubbish before free
	#endif	

	DisposePtr((Ptr)pointer); /// Free memory
}

void DestroyHandle(void** handle)
{
	#if Project_Under_Development && Track_Memory_Allocations
		UnregisterPointer(pointer, '_hdl');
	#endif			
			
	#if Project_Under_Development && Initialise_Allocated_Memory
		InitialiseMemory((char*)*handle, GetHandleSize((Handle)handle)); /// Fill with rubbish before free
	#endif	

	DisposeHandle((Handle)handle); /// Free memory
}

#if Project_Under_Development

#if Assert_Memory_Locking

void LockHandleAssert(void** handle)
{
	SignedByte	statusBefore, statusAfter;
	
	Assert(handle);
	
	statusBefore=HGetState((Handle)handle); /// Check handle's status before locking
	HLock((Handle)handle); /// Actually lock handle
	statusAfter=HGetState((Handle)handle); /// Check handle's status after locking
	
	Assert(statusBefore!=statusAfter); /// If it was locked before, these would be =
}

void UnlockHandleAssert(void** handle)
{
	SignedByte	statusBefore, statusAfter;
	
	Assert(handle);

	statusBefore=HGetState((Handle)handle); /// Check handle's status before unlocking
	HUnlock((Handle)handle); /// Actually unlock handle
	statusAfter=HGetState((Handle)handle); /// Check handle's status after unlocking
	
	Assert(statusBefore!=statusAfter); /// If it was unlocked before, these would be =
}
	
#endif

#if Initialise_Allocated_Memory

void InitialiseMemory(char* pointer, Size length)
{
	char	*trasher, *topAddress;
	
	trasher=pointer; /// Start of trashing address
	topAddress=trasher+length; /// End of trashing address
	
	while (trasher<topAddress)
		*trasher++=(char)0xAA; /// Fill it up with those 0xAA bytes
}
	
#endif

#endif

void GlueSetTheCursorToArrow()
{
	Cursor		arrowCursor;
	
	SetCursor(GetQDGlobalsArrow(&arrowCursor));
}

Size GlueGetScrap(Handle handle, OSType type)
{
	OSStatus			error;
	ScrapRef			scrapRef;
	ScrapFlavorFlags	flavorFlags;
	Size				scrapSize;
	SInt8				handleState;
	
	Assert((!handle)||(*handle));
	
	error=GetCurrentScrap(&scrapRef);
		_i(error)
		
	error=GetScrapFlavorFlags(scrapRef, type, &flavorFlags); /// similar API to drag manager
		_i(error)
		
	error=GetScrapFlavorSize(scrapRef, type, &scrapSize);
		_i(error)
	
	if (handle) { /// we can take a zero handle just to check the data size, like Classic call
		SetHandleSize(handle, scrapSize);
		error=MemError(); /// check we were able to resize handle
			_i(error)
			
		handleState=HGetState(handle); /// to allow restoring whatever state it was in
		HLock(handle);
		error=GetScrapFlavorData(scrapRef, type, &scrapSize, *handle);
		HSetState(handle, handleState);
			_i(error)
	}
		
	return scrapSize;
	
	_e
	if (error<0) /// according to Classic scrap API, should return zero or less to indicate error
		return error;
	else
		return 0;
}

void SetItemHandle(DialogRef dialog, short item, Handle handle)
{
	Handle		oldHandle;
	short		type; // dummy variable
	Rect		rect; // dummy variable

	Assert(dialog);

	GetDialogItem(dialog, item, &type, &oldHandle, &rect);
	SetDialogItem(dialog, item, type, handle, &rect);
}


Handle GetItemHandle(DialogRef dialog, short item)
{
	Handle		handle;
	short		type; // dummy variable
	Rect		rect; // dummy variable
	
	Assert(dialog);

	GetDialogItem(dialog, item, &type, &handle, &rect);
	return handle;
}

void GetItemRect(DialogRef dialog, short item, Rect* rect)
{
	Handle		handle; // dummy variable
	short		type; // dummy variable
	
	Assert(dialog);

	GetDialogItem(dialog, item, &type, &handle, rect);
}

short GetItemValue(DialogRef dialog, short item)
{
	Assert(dialog);
	Assert(GetItemHandle(dialog, item));

	return GetControlValue((ControlHandle)GetItemHandle(dialog, item));
}

void SetItemValue(DialogRef dialog, short item, short value)
{
	Assert(dialog);
	Assert(GetItemHandle(dialog, item));

	SetControlValue((ControlHandle)GetItemHandle(dialog, item), value);
}

void SetItemEnabled(DialogRef dialog, short item, Boolean enabled)
{
	Assert(dialog);
	Assert(GetItemHandle(dialog, item));

	Assert(GetControlOwner((ControlHandle)GetItemHandle(dialog, item))==GetDialogWindow(dialog));
		
	HiliteControl((ControlHandle)GetItemHandle(dialog, item),
	              enabled ? (short)0 : (short)255); // 255 disables it
}

Boolean ToggleItem(DialogRef dialog, short item)
{
	Boolean		value;
	
	Assert(dialog);

	value=!GetItemValue(dialog, item); // flip value
	SetItemValue(dialog, item, value); // set to flipped value
	return value; // return flipped value
}

void GetItemText(DialogRef dialog, short item, Str255 string)
{
	Assert(dialog);
	Assert(GetItemHandle(dialog, item));

	GetDialogItemText(GetItemHandle(dialog, item), string);
}

long GetItemNumber(DialogRef dialog, short item)
{
	Str255		string;
	long		number;

	Assert(dialog);
		
	GetItemText(dialog, item, string); // gets text
	StringToNum(string, &number); // converts to number
	return number; // returns number
}

void SetItemText(DialogRef dialog, short item, const Str255 string)
{
	Assert(dialog);
	Assert(GetItemHandle(dialog, item));

	SetDialogItemText(GetItemHandle(dialog, item), string);
}

void SetItemNumber(DialogRef dialog, short item, long number)
{
	Str255		string;
	
	Assert(dialog);
	
	NumToString(number, string); // converts to text
	SetItemText(dialog, item, string); // sets text 
}

Error FindApplication(OSType creator, ProcessSerialNumber *process)
{
	Error		error;
	FSSpec		fileSpec;

	Assert(process);
	
	error=FindApplicationProcess(creator, process); // see if it's already running
		_i(error==0)
	error=FindApplicationFile(creator, &fileSpec); // if not, look for it
		_i(error)
	error=OpenApplicationFile(&fileSpec, process); // if it was found, run it
		
	_e
	return error;
}

Error FindApplicationProcess(OSType creator, ProcessSerialNumber *process)
{
	Error				error;
	ProcessInfoRec		infoRec;
	Str255				processName;
	FSSpec				processFile;

	Assert(process);

	process->highLongOfPSN=0; // so GetNextProcess retries first one
	process->lowLongOfPSN=kNoProcess;
	infoRec.processInfoLength=sizeof(infoRec); // param block vagaries
	infoRec.processName=processName;
	infoRec.processAppSpec=&processFile;
	
	while (1) {
		error=GetNextProcess(process); // Get the next process
			_i(error); // if none left, we've failed to find it
		error=GetProcessInformation(process, &infoRec); // find out about it
			_i(error);
			
		if (infoRec.processType=='APPL' && infoRec.processSignature==creator)
			break; // check it's an application and has the right creator
	}
	
	_e
	return error;
}

Error FindApplicationFile(OSType creator, FSSpec* fileSpec)
{
	Error			error=0;
	HVolumeParam	volume;
	DTPBRec			desktop;
	FInfo			fileInfo;

	#if Using_Carbon_API
		long			gestaltResponse;
		FSRef			fileRef;
	#endif
	
	Assert(fileSpec);

	#if Using_Carbon_API
		if (!Gestalt(gestaltSystemVersion, &gestaltResponse)&&(gestaltResponse>=0x1000)) {
			error=LSFindApplicationForInfo(creator, NULL, NULL, &fileRef, NULL);
				_i2(error)
				
			error=FSGetCatalogInfo(&fileRef, kFSCatInfoNone, NULL, NULL, fileSpec, NULL);
				_i2(error)
			
			_g // we found it so return it
		}
	#endif
	
	_e2
	
	volume.ioVolIndex=0;
	while (1) {	
		volume.ioVolIndex++; // increment required volume
		volume.ioNamePtr=0; // don't want to retrieve name
		error=PBHGetVInfoSync((HParamBlockRec*)&volume); // get volume's refnum
			_i(error);

		desktop.ioVRefNum=volume.ioVRefNum; // now use volume's ref num
		desktop.ioNamePtr=0;
		desktop.ioIndex=0; // get main desktop file
		
		if (!PBDTGetPath(&desktop)) { // try to get desktop file path
			desktop.ioFileCreator=creator;
			desktop.ioNamePtr=fileSpec->name;
			desktop.ioIndex=1; // find first one first
			
			while (!PBDTGetAPPLSync(&desktop)) { // look for required application
				fileSpec->vRefNum=volume.ioVRefNum;
				fileSpec->parID=desktop.ioAPPLParID;
				if (!FSpGetFInfo(fileSpec, &fileInfo) && (fileInfo.fdCreator==creator))
					_g; // otherwise it could be old in desktop database
				desktop.ioIndex++; // try for next application instance
			}
		}
	}

	_e
	return error; // will be not found if none found
}

Error OpenApplicationFile(FSSpec* fileSpec, ProcessSerialNumber *process)
{
	Error				error;
	LaunchParamBlockRec	launch;
	
	Assert(fileSpec);
	Assert(process);

	launch.launchBlockID=extendedBlock; // param block vagaries
	launch.launchEPBLength=extendedBlockLen;
	launch.launchFileFlags=0;
	launch.launchControlFlags=launchContinue|launchNoFileFlags; // DO switch
	launch.launchAppSpec=fileSpec;
	launch.launchAppParameters=0;
	
	error=LaunchApplication(&launch); // call Mac OS launch application
		TestError(error);
		_i(error);
		
	*process=launch.launchProcessSN; // retrieve resultant process serial number
	
	_e
	return error;
}	

Error CreateMemoryBuffer(MemoryBuffer* buffer)
{
	Error		error;
	
	Assert(buffer);
	
	buffer->handle=(Handle)AllocHandle(&error, 0); // start with empty handle
		TestError(error);
		_i(error);
		
	buffer->allocated=0; // initialise counters
	buffer->used=0;

	_e
	return error;
}

Error WriteMemoryBuffer(MemoryBuffer* buffer, const void* dataPtr, long count)
{
	Error	error;
	long	needed;

	Assert(buffer);
	Assert(dataPtr||!count);

	needed=buffer->used+count; // how much we will need in total
	if (needed>buffer->allocated) { // time to allocate some more memory
		needed=((needed/memoryBufferBlockSize)+1)*memoryBufferBlockSize;
			// round up to the nearest whole block
		SetHandleSize(buffer->handle, needed); // resize handle
			error=TestMemError(buffer->handle);
			_i(error);
			
		buffer->allocated=needed;

	} else
		error=0; // otherwise error would be undefined

	BlockMove(dataPtr, *(buffer->handle)+buffer->used, count); // move data in
	buffer->used+=count; // increment the amount used
	Assert(buffer->used<=buffer->allocated); // just in case something went awry!
	
	_e
	return error;
}

Handle CloseMemoryBuffer(MemoryBuffer* buffer)
{
	Assert(buffer);

	SetHandleSize(buffer->handle, buffer->used); // shrink buffer down to amount used
	
	return buffer->handle; // doesn't have to do anything much to close buffer
}